/*
Copyright 2013 Paul Soulsby www.soulsbysynths.com
This file is part of Atmegatron.

Atmegatron is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Atmegatron is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Atmegatron.  If not, see <http://www.gnu.org/licenses/>.
*/
/*
USE_GITHUB_USERNAME=soulsbysynths
*/
//***************ATMEGATRON MASTER CODE***********************

#ifndef cbi
#define cbi(sfr, bit) (_SFR_BYTE(sfr) &= ~_BV(bit))
#endif
#ifndef sbi
#define sbi(sfr, bit) (_SFR_BYTE(sfr) |= _BV(bit))
#endif

#define MEM_PATCHSIZE 21    //size of flash mem patch
#define CTRL_FILT 0         //digital control nums
#define CTRL_Q 1
#define CTRL_ENV 2
#define CTRL_LFO 3
#define CTRL_AMP 4
#define CTRL_FX 5
#define FUNC_WAVEA 0        //function nums
#define FUNC_WAVEB 1
#define FUNC_FENVA 2
#define FUNC_FENVDR 3
#define FUNC_FENVS 4
#define FUNC_AENVA 5
#define FUNC_AENVD 6
#define FUNC_AENVS 7
#define FUNC_AENVR 8
#define FUNC_LFOTYPE 9
#define FUNC_LFOSPEED 10
#define FUNC_ARPTYPE 11
#define FUNC_ARPSPEED 12
#define FUNC_PORTA 13
#define FUNC_BITCRUSH 14
#define FUNC_MEM 15
#define WAVE_LEN 32        //length of wavetable
#define POLL_HARD 1        //poll the function/val dials and digital controls
#define POLL_MIDI 1        //poll the midi input
#define POLL_TEST 0        //test mode
#define MASTER_BEATTICK 500//length of beat in ticks. 500 = 120bpm i.e. 1 beat = 500ms (this is overridden when MIDI clock received)
#define MICROSECONDS_PER_TIMER2_OVERFLOW 32

unsigned long master_tick = 0;                        //current tick.  used for all timing of lfos, arp and envs
unsigned int master_frac = 0;
volatile uint16_t master_index[2] = {0,0};               //position in master_output wavetable
volatile char master_output[2][WAVE_LEN];   //output wavetable
volatile unsigned long master_ocr[2] = {142,1136};        //interrupt frequency (sets pitch of synth)
volatile byte master_prescale = 2;
unsigned long master_sf[2] = {14080,14080};                   //sample frequency, ocr1 derived from this
volatile byte master_oscmix[2] = {0,0};
volatile boolean master_oscmute[2] = {false,false};
byte oscNum = 0;
	
//Setup.  This runs once on bootup
void setup() {

	Init_Register();             //Setup PWM pins on Atmega.  Also sets faster ADC read times.
	Init_Waves();                //initialise wavetables
	Init_Hardware();             //initialise input controls and output leds
	
	
	LFO_CalcVal();               //initialise lfo
	AEnv_CalcVal();              //initialise amplitdue envelope
	FEnv_CalcVal();              //initialise filter/pitch envelope
	Filt_CalcVals(0);             //initialise filter
	Filt_CalcVals(1);
	
	if (POLL_MIDI == 1){         //open midi port (standard Arduino serial port)
		MIDI_Init();
	}
	else{
		Serial.begin(9600);       //open the port for debugging in Arduino enviroment
	}

	Memory_Vanilla();                //load the Vanilla patch
	Hardware_LED_StartFlash(0,5);    //flash to say "I'm ready!"
}

//Loop.  This loops forever, after setup is complete.  It controls all tasks except the audio output.
void loop(){

	oscNum = 1-oscNum;
	if (POLL_MIDI==1){
		MIDI_Poll();                                    //poll the MIDI input
	}
	if (POLL_HARD==1){
		Hardware_Poll();                                //poll the hardware (potentiometers, rotary encoders and buttons)
	}
	if (POLL_TEST==1){
		if (master_tick>=Testmode_Get_NextTick()){      //if test mode is being used, trigger next test note if it's time to
			Testmode_ProcessNext();
		}
	}
	if(oscNum==0){
		Arp_Poll();                                       //poll the arpeggiator and see if it's time for a note (if it's on)
		LFO_CalcVal();                                    //calculate the current output of the LFO
		AEnv_CalcVal();                                   //calculate the current output of the amplitude envelope
		FEnv_CalcVal();                                   //calculate the current output of the filter/pitch envelope
	}
	Pitch_Process(oscNum);                                  //calculate the current pitch
	Filt_CalcVals(oscNum);                                  //calculate the variables used by the filter
	Master_Let_SampleFreq(oscNum);                          //set interrupt freq based on above calculation
	if (Aenv_Get_State()>0){                          //if a key is down (actually is amplitude envelope idle):
		Wave_Process(oscNum);                                 //process the wavetable (filter/envelope/fx etc...)
	}
	else{                                             //otherwise clear the wavetable.  This is to stop
		Wave_Clear();
	}
}

void Master_Let_SampleFreq(byte oscNum){                       //set the sample frequency (pitch)
	unsigned long tmp;
	if (Pitch_Get_PitchChanged(oscNum)==true){              //has the pitch actually changed?
		master_sf[oscNum] = Pitch_Get_FreqCalc(oscNum)*WAVE_LEN;      //calc sample freq
		tmp = F_CPU/master_sf[oscNum];
		if (tmp>65535){
			tmp = 65535;
			master_oscmute[oscNum] = true;
		}
		else if (tmp < 238){
			tmp = 238;
			master_oscmute[oscNum] = true;
		}
		else{
			master_oscmute[oscNum] = false;
		}
		if(oscNum==1){
			master_ocr[1] = tmp;                  //convert to interrupt time
		}
		else{
			if (tmp>16383){
				master_ocr[0] = tmp >> 8;
				master_prescale = 4;
			}
			else if (tmp>2047){
				master_ocr[0] = tmp >> 6;
				master_prescale = 3;
			}
			else if (tmp>255){
				master_ocr[0] = tmp >> 3;
				master_prescale = 2;
			}
			else{
				master_ocr[0] = tmp;
				master_prescale = 1;
			}
		}
	}
}
unsigned long Master_Get_SampleFreq(byte oscNum){
	return master_sf[oscNum];
}


//Interrupt loop.  Sets PWM output (i.e. generates the audio)
ISR(TIMER2_OVF_vect) {
	int master_out = 0;
	byte i;
	unsigned long m = master_tick;
	unsigned int f = master_frac;
	
	f += MICROSECONDS_PER_TIMER2_OVERFLOW;

	if (f >= 1000) {
		f -= 1000;
		m += 1;
	}
	
	master_tick = m;
	master_frac = f;

	for (i=0;i<2;i++){
		if(master_oscmute[i]==false){
			master_out += master_output[i][master_index[i]] >> master_oscmix[i];
		}
	}
	master_out = master_out >> 1;
	
	if (master_out>127){
		OCR2B = 255;
	}
	else if (master_out<-128){
		OCR2B = 0;
	}
	else{
		OCR2B = master_out + 128;
	}

}

//Interrupt loop.  Sets PWM output (i.e. generates the audio)
ISR(TIMER1_COMPA_vect) {
	if (master_index[1] >= WAVE_LEN - 1) {                //if at end of wavetable, go back to start
		master_index[1] = 0;
	}
	else{
		master_index[1]++;
	}
	//OCR2B = master_output[master_index];              //sets sample value
	//increment position in wavetable
	OCR1A = master_ocr[1];                              //sets pitch
}


//Interrupt loop.  Sets PWM output (i.e. generates the audio)
ISR(TIMER0_COMPA_vect) {
	if (master_index[0] >= WAVE_LEN - 1) {                //if at end of wavetable, go back to start
		master_index[0] = 0;
	}
	else{
		master_index[0]++;
	}
	switch (master_prescale){
		case 1:
			sbi(TCCR0B,CS00);
			cbi(TCCR0B,CS01);
			cbi(TCCR0B,CS02);
			break;
		case 2:
			cbi(TCCR0B,CS00);
			sbi(TCCR0B,CS01);
			cbi(TCCR0B,CS02);
			break;
		case 3:
			sbi(TCCR0B,CS00);
			sbi(TCCR0B,CS01);
			cbi(TCCR0B,CS02);
			break;
		case 4:
			cbi(TCCR0B,CS00);
			cbi(TCCR0B,CS01);
			sbi(TCCR0B,CS02);
		break;
	}
	OCR0A = master_ocr[0];                              //sets pitch
	
}




