/*
Copyright 2013 Paul Soulsby www.soulsbysynths.com
    This file is part of Atmegatron.

    Atmegatron is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Atmegatron is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Atmegatron.  If not, see <http://www.gnu.org/licenses/>.
*/
//frequency of all midi notes  (rounded to nearest integer)
prog_uint16_t  MIDI_freqs[128] PROGMEM = {8,9,9,10,10,11,12,12,13,14,15,15,16,17,18,19,21,22,23,24,26,28,29,31,33,35,37,39,41,44,46,49,52,55,58,62,65,69,73,78,82,87,92,98,104,110,117,123,131,139,147,156,165,175,185,
196,208,220,233,247,262,277,294,311,330,349,370,392,415,440,466,494,523,554,587,622,659,698,740,784,831,880,932,988,1047,1109,1175,1245,1319,1397,1480,1568,1661,1760,1865,1976,2093,2217,2349,2489,2637,2794,
2960,3136,3322,3520,3729,3951,4186,4435,4699,4978,5274,5588,5920,6272,6645,7040,7459,7902,8372,8870,9397,9956,10548,11175,11840,12544};								

//midi constants
#define NOTE_OFF 128
#define NOTE_ON 144
#define PITCH_WHEEL 224
#define CONTROLLER 176
#define SYSEXBEGIN 240
#define SYSEXEND 247
#define CC_PITCHLFO 1
#define CC_PORTAMENTO 5
#define CC_FILTERENV 16
#define CC_DISTORTION 17
#define CC_FILTCUTOFF 74 
#define CC_AMPENVR 72
#define CC_AMPENVA 73
#define CC_FILTRES 71
#define CC_AMPENVD 75
#define CC_LFOCLOCKDIV 79
#define CC_PWM 91
#define CC_AMPLFO 92
#define CC_FILTLFO 93
#define CC_PITCHENV 94
#define CC_FLANGE 95
#define CC_ALLNOTESOFF 123


//local vars

boolean arp_triggernext=false;                //trigger arpeggiator note on next midi clock

//lets and gets
byte MIDI_curnote = 57;                       //current note pressed
byte MIDI_curbutnote = 60;                    //current note of test button
boolean MIDI_keydown[128];                    //store which keys are down for arpeggiator
char MIDI_totnoteson =0;                      //total notes held down 
char MIDI_lasttotnoteson = 0;                 //last total notes held down
boolean MIDI_clockpresent=false;              //is midi clock present?
unsigned long MIDI_clocktick=0;               
unsigned long MIDI_clockticksperarpstep=12;
unsigned long MIDI_clocknextarp=0;
char MIDI_pbend_level=0; 
byte MIDI_channel=0;

//Lets and Gets
byte MIDI_Get_curNote(){                            //get current midi note
  return MIDI_curnote;
}
unsigned int MIDI_Get_Freq(byte notenum){          //get current midi frequency (potentially different to actual current frequency in pitch tab)
  return pgm_read_word(&(MIDI_freqs[notenum]));
}
boolean MIDI_Get_KeyDown(byte notenum){            //find out if note is down
  return MIDI_keydown[notenum];
}

void MIDI_Let_ClockArpSpeed(unsigned int newspeed){                            //set clock arp speed.  not actual speed.  see Arp_Let_Speed for how this works
  MIDI_clockticksperarpstep = ((unsigned long)newspeed+1) * 24 >> 13;
}


boolean MIDI_Get_ClockPresent(){                   //return if midi clock is preset
  return MIDI_clockpresent;
}
byte MIDI_Set_Channel(byte newchannel){
  if(newchannel!=MIDI_channel){
    MIDI_channel = newchannel;
    Memory_Channel_Write(newchannel);
  }
      
}

//MIDI meat
//initialise serial port
void MIDI_Init(){                                  //enable midi.  just uses standard arduino serial ports.
  MIDI_channel = Memory_Channel_Read();
  Serial.begin(31250);
}
//********MIDI note stuff***********
//press key
void MIDI_NoteOn(byte notenum){
  MIDI_curnote = notenum;                                    //set current note = new note
  MIDI_lasttotnoteson = MIDI_totnoteson;                     //set last total notes on
  MIDI_totnoteson++;                                         //increment total notes on
  MIDI_keydown[notenum] = true;                              //set midi key down = true
  if (Arp_Get_Type()==0){                                    //if arpeggiator off then
    MIDI_TriggerNote(notenum);                               //trigger note
  }
  else if (MIDI_lasttotnoteson==0 && MIDI_totnoteson==1) {   //if arpeggiator on and this is the first note pressed down
    Arp_Reset();                                             //reset arpeggiator and lfo, so they sync together       
    LFO_Reset();          
  }
}
//release key
void MIDI_NoteOff(byte notenum){
  MIDI_keydown[notenum] = false;                             //set midi key down = false
  MIDI_lasttotnoteson = MIDI_totnoteson;                     //set last total notes on
  MIDI_totnoteson--;                                         //decrement total notes on
  if (MIDI_totnoteson<=0){                                   //if total notes on = 0 (ie all keys released)  (use <= for safety)
    AEnv_Release();                                          //release amplitude envelope
    FEnv_Release();                                          //release filter envelope
    if (MIDI_totnoteson<0){                                  //safety - if tot notes on < 0
      MIDI_totnoteson = 0;                                   //reset to 0 (<0 not possible)
    }
  }
}
//*********TRIGGER A NOTE**************** (used by arpeggiator too)
void MIDI_TriggerNote(byte notenum){                                      
  Pitch_Let_NextFreq(pgm_read_word(&(MIDI_freqs[notenum])));                //set next frequency (pitch)
  Pitch_ResetPorta();                                                       //set portamento start tick
  AEnv_Trigger();                                                           //trigger amplitude envelope
  FEnv_Trigger();                                                           //trigger filter envelope
}
//reset all notes
void MIDI_Reset(){
  byte i;
  for (i=0;i<=127;i++){
    MIDI_NoteOff(i);
  }
}

//press the test button (value knob)
void MIDI_TestButtonDown(){                                                
  Pitch_Let_NextFreq(pgm_read_word(&(MIDI_freqs[MIDI_curbutnote])));       //set next frequency
  Pitch_ResetPorta();                                                      //set portamento start tick
  AEnv_Trigger();                                                          //trigger amplitude envelope
  FEnv_Trigger();                                                          //trigger filter envelope
}
//release the test button (value knob)
void MIDI_TestButtonUp(){                                                  
  if (MIDI_totnoteson<=0){                                                 //prevent stuck notes
    AEnv_Release();                                                        //release amplitude envelope
    FEnv_Release();                                                        //release filter envelope
  }  
}
//increment test button note (clockwise turn)
void MIDI_TestButtonInc(){                                                 
  if (MIDI_curbutnote<126){                                                //if cur test button note < max midi note
    MIDI_curbutnote++;                                                     //increment
    Pitch_Let_NextFreq(pgm_read_word(&(MIDI_freqs[MIDI_curbutnote])));     //set next frequency (pitch)
  }
}
//decrement test button note (anti clockwise turn)
void MIDI_TestButtonDec(){                                                 
  if (MIDI_curbutnote>1){                                                  //if cur test button note > min midi note
    MIDI_curbutnote--;                                                     //decrement
    Pitch_Let_NextFreq(pgm_read_word(&(MIDI_freqs[MIDI_curbutnote])));     //set next frequency (pitch)
  }
}

//*******MIDI clock stuff********
//MIDI clock start received
void MIDI_ClockStart(){
  MIDI_clockpresent = true;       //set clock present 
  MIDI_clocktick = 0;             //reset clock tick counter
  MIDI_clocknextarp = 0;          //reset next arpeggiator note tick
  if (Arp_Get_Type()!=0) {        //if arpeggiator on
    Arp_Reset();                  //reset it (so synced with clock)
  }
  LFO_Reset();                    //reset lfo (so synced with clock)
}
//No implementation of clock continue at present.  just does clock start
void MIDI_ClockContinue(){
   MIDI_ClockStart();
}
//MIDI clock stop received
void MIDI_ClockStop(){
  MIDI_clockpresent = false;       //set clock not present
}
//MIDI clock tick received
void MIDI_ClockTick(){
  if (MIDI_clockpresent==true) {                                  //check clock has acutally been started (not rogue data)
    if (MIDI_clocktick>=MIDI_clocknextarp && Arp_Get_Type()!=0){  //if midi clock counter > midi clock value for next arppeggiator trigger (and arpeggitor on) :
      arp_triggernext = true;                                     //trigger arpeggiator at end of current midi receive (bottom of MIDI_poll)
    }
    if (MIDI_clocktick%24 == 0){                                  //every 24 clock ticks (every beat):
      Hardware_LED_StartFlash(1,1);                               //flash the midi led
    }
    MIDI_clocktick++;                                             //increment midi clock counter
  }
  
}
//get midi clock counter
unsigned long MIDI_Get_ClockTick(){
  return MIDI_clocktick;
}

//get pitch bend  (only using msb of pitch bend data)
char MIDI_Get_PitchBend_Level(){
  return MIDI_pbend_level;
}



//**************MIDI POLL ********************* 
void MIDI_Poll(){
  byte incomingByte=0;   
  static byte notebyte=0;        //these are all static because there's a small chance that the subroutine is exited before the 2 bytes of MIDI needed for note_on are received.
  static byte velocitybyte=0;
  static byte statusbuffer=0;
  static boolean firstbyte;
  if (Serial.available() > 0) {                    //if there is something in the serial input buffer
    do {
      incomingByte = Serial.read();                // read the incoming byte:
      if (incomingByte>247) {                      // ****this is where MIDI clock stuff is done***
        switch (incomingByte){
          case 248:                                //MIDI clock tick
            MIDI_ClockTick();
            break;
          case 250:                                //MIDI clock start
            MIDI_ClockStart();
            break;
          case 251:                                //MIDI clock continue (not implemented)
            MIDI_ClockContinue();
            break;
          case 252:                                //MIDI clock stop
            MIDI_ClockStop();
            break;
        }
      }

      else if (incomingByte>127) {                 //*****this is where all other message types are done ******
        statusbuffer = incomingByte;               //status buffer stores what kind of buffer it is
        firstbyte = true;                          //it must be the first byte of pair
        notebyte = 0;                              //so reset the note byte
        velocitybyte = 0;                          //and velocity byte, ready incase it's a note
      }
      else if (statusbuffer!=0) {                  //******all other message data*******
        if (firstbyte == true) {                   //if it's first byte of pair
          notebyte = incomingByte;                 //set note = incoming byte
          firstbyte = false;                       //and set = false, so ready for second byte
        }
        else {                                     // so must be second byte of pair then
          velocitybyte = incomingByte;                                                            //set velocitybyte as incoming byte (even if not note data)
          if (statusbuffer == (NOTE_ON | MIDI_channel) && velocitybyte != 0) {                                     //is it a note on?
            Hardware_LED_StartFlash(1,1);                                                         //flash midi led
            MIDI_NoteOn(notebyte);                                                                //trigger note
          }
          else if (statusbuffer == (NOTE_OFF | MIDI_channel) || (statusbuffer == (NOTE_ON | MIDI_channel) && velocitybyte == 0)) {  //is it a note off?  (there's 2 ways to trigger note off:  note_off or Note_on with velocity =0)
            Hardware_LED_StartFlash(1,1);                                                         //flash midi led
            MIDI_NoteOff(notebyte);                                                               //release note
          }
          else if (statusbuffer == (PITCH_WHEEL | MIDI_channel)){                                                  //is it pitch wheel?
            MIDI_pbend_level = velocitybyte - 64;                                                 //set pbend (may aswell do the offset here).  only msb of pitch bend used
          }
          else if (statusbuffer == (CONTROLLER | MIDI_channel)){                                                   //is it a controller message?  I have attempted to use controller numbers appoximating the spec of official controller list.  Used "MISC" controllers where necessary
            switch (notebyte) {
              case CC_PITCHLFO:                                                                   //pitch lfo (mod wheel).  This is just an override of the pitch lfo knob
                Hardware_Let_Ctrl(1,CTRL_LFO,velocitybyte >> 1);                                  //default uses >>1 for typical mod wheel vibrato.  use  << 1 for full range (default full range = 1 octave)
                break;
              case CC_PORTAMENTO:                                                                 //portamento
                Hardware_Let_Value(FUNC_PORTA,velocitybyte >> 3);                                 //scale to values 0-15
                break;
              case CC_FILTERENV:                                                                  //filter envelope
                Hardware_Let_Ctrl(0,CTRL_ENV,velocitybyte << 1);                                  //scale to values 0-255
                break;
              case CC_DISTORTION:                                                                 //distortion
                Hardware_Let_Ctrl(0,CTRL_FX,velocitybyte << 1);                                   //scale to values 0-255
                break;
              case CC_FILTCUTOFF:                                                                 //filter cutoff frequency
                Hardware_Let_Ctrl(0,CTRL_FILT,velocitybyte << 1);                                 //scale to values 0-255
                break;
              case CC_AMPENVR:                                                                    //amplitude envelope release
                Hardware_Let_Value(FUNC_AENVR,velocitybyte >> 3);                                 //scale to values 0-15
                break;
              case CC_AMPENVA:                                                                    //amplitude envelope attack
                Hardware_Let_Value(FUNC_AENVA,velocitybyte >> 3);                                 //scale to values 0-15
                break;
              case CC_FILTRES:                                                                    //filter resonance
                Hardware_Let_Ctrl(0,CTRL_Q,velocitybyte << 1);                                    //scale to values 0-255
                break;
              case CC_AMPENVD:                                                                    //amplitude envelope decay
                Hardware_Let_Value(FUNC_AENVD,velocitybyte >> 3);                                 //scale to values 0-15
                break;
              case CC_LFOCLOCKDIV:                                                                //lfo rate (speed).    ******DUBSTEP PRODUCERS THIS IS CONTROLLER 79!!!********
                Hardware_Let_Value(FUNC_LFOSPEED,velocitybyte >> 3);                              //scale to values 0-15 
                break;
              case CC_PWM:                                                                        //pulse width modulation
                Hardware_Let_Ctrl(1,CTRL_AMP,velocitybyte << 1);                                  //scale to values (0-255)
                break;
              case CC_AMPLFO:                                                                     //amplitude lfo
                Hardware_Let_Ctrl(0,CTRL_AMP,velocitybyte << 1);                                  //scale to values (0-255)
                break;
              case CC_FILTLFO:                                                                    //filter lfo
                Hardware_Let_Ctrl(0,CTRL_LFO,velocitybyte << 1);
                break;
              case CC_PITCHENV:                                                                   //pitch envelope
                Hardware_Let_Ctrl(1,CTRL_ENV,velocitybyte << 1);
                break;
              case CC_FLANGE:                                                                     //phaser
                Hardware_Let_Ctrl(1,CTRL_FX,velocitybyte << 1);
                break;
              case CC_ALLNOTESOFF:
                MIDI_Reset();
            }
          }                                                                       
          notebyte = 0;                                                                           //now reset them ready for next note
          velocitybyte = 0; 
          firstbyte = true;         
        }
      }
    } while (Serial.available() > 0);                                                             //loop until serial buffer is empty

    if (arp_triggernext==true && MIDI_totnoteson>0){                                              //midi clock tick was received and it's time to fire an arpeggiator note!
      do{                                                                                         //nasty  do loop.  this is important, incase midi clock ticks are faster than Atmegatron polling loop and multiple notes need to be triggered now
        Arp_TriggerStep();                                                                        //trigger arpeggiator
        MIDI_clocknextarp += MIDI_clockticksperarpstep;                                           //MIDI_clockticksperarpstep  MUST never be 0
      } while (MIDI_clocknextarp<MIDI_clocktick);                                                 //loop until all notes triggered that need to be are
      arp_triggernext = false;                                                                    //triggered all notes, so set false
    }
  } 
}

