/*
Copyright 2013 Paul Soulsby www.soulsbysynths.com
This file is part of Atmegatron.

Atmegatron is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Atmegatron is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Atmegatron.  If not, see <http://www.gnu.org/licenses/>.
*/

//******Read and write patches to flash memory**************

#include <EEPROM.h>                          //library comes with arduino software
#define SYSEXBEGIN 240
#define SYSEXEND 247

//save patch
void Memory_Save(byte patchnum){
	byte i, v, lsb, msb, anal, out;
	int addr;
	boolean b;
	
	cli();                                    //halt interrupts (i.e. audio engine).  don't want it interferring with eeprom writing
	
	addr = patchnum * MEM_PATCHSIZE;          //find start address  (normal patch size = 19 bytes)
	v = 0;
	
	for (i=0;i<8;i++){                        //writes all 16 function values to memory
		lsb = Hardware_Get_Value(v);            //each function is value 0-15 = 4 bit value.
		msb = Hardware_Get_Value(v+1);          //so can store 2 function values in each byte (8 bits)
		msb = msb << 4;                         //this is done by bit-shifting every other value left 4 bits
		out = msb | lsb;                        //and ORing with other value
		EEPROM.write(addr+i, out);              //then write to eeprom
		v += 2;
	}
	//write all ctrl values to memory
	anal = Hardware_Get_Ctrl(0, CTRL_FILT);   //first the 6 red bank ctrls
	EEPROM.write(addr+8, anal);
	anal = Hardware_Get_Ctrl(0, CTRL_Q);
	EEPROM.write(addr+9, anal);
	anal = Hardware_Get_Ctrl(0, CTRL_ENV);
	EEPROM.write(addr+10, anal);
	anal = Hardware_Get_Ctrl(0, CTRL_LFO);
	EEPROM.write(addr+11, anal);
	anal = Hardware_Get_Ctrl(0, CTRL_AMP);
	EEPROM.write(addr+12, anal);
	anal = Hardware_Get_Ctrl(0, CTRL_FX);
	EEPROM.write(addr+13, anal);
	anal = Hardware_Get_Ctrl(1, CTRL_FILT);   //first the 6 red bank ctrls
	EEPROM.write(addr+14, anal);
	anal = Hardware_Get_Ctrl(1, CTRL_Q);
	EEPROM.write(addr+15, anal);
	anal = Hardware_Get_Ctrl(1, CTRL_ENV);    //then the 4 green bank ctrls
	EEPROM.write(addr+16, anal);
	anal = Hardware_Get_Ctrl(1, CTRL_LFO);
	EEPROM.write(addr+17, anal);
	anal = Hardware_Get_Ctrl(1, CTRL_AMP);
	EEPROM.write(addr+18, anal);
	anal = Hardware_Get_Ctrl(1, CTRL_FX);
	EEPROM.write(addr+19, anal);
	
	//write function shift modes to memory.
	out = 0;                                  //these are boolean, so 1 bit each.  so 8 can be store in each byte (i.e. 8 bits).  out is used to build up output byte
	if (Hard_Get_Shift(FUNC_WAVEA)==true){     //get each shift mode and set/clear each bit of out.   First wavetable bank
		bitSet(out, 0);
	}
	else{
		bitClear(out, 0);
	}
	if (Hard_Get_Shift(FUNC_WAVEB)==true){    //filter normalise mode
		bitSet(out, 1);
	}
	else{
		bitClear(out, 1);
	}
	if (Hard_Get_Shift(FUNC_FENVA)==true){   //filt/pitch env invert mode
		bitSet(out, 2);
	}
	else{
		bitClear(out, 2);
	}
	if (Hard_Get_Shift(FUNC_LFOTYPE)==true){  //lfo invert mode
		bitSet(out, 3);
	}
	else{
		bitClear(out, 3);
	}
	if (Hard_Get_Shift(FUNC_ARPTYPE)==true){  //arp ping-pong mode
		bitSet(out, 4);
	}
	else{
		bitClear(out, 4);
	}
	if (Hard_Get_Shift(FUNC_PORTA)==true){    //proportional portamento
		bitSet(out, 5);
	}
	else{
		bitClear(out, 5);
	}
	if (Hard_Get_Shift(FUNC_BITCRUSH)==true){  //bitcrush before filter mode
		bitSet(out, 6);
	}
	else{
		bitClear(out, 6);
	}
	EEPROM.write(addr+20, out);                //write compiled bits to eeprom
	sei();                                     //restart interrupts (audio engine)
	
}

//load patch
void Memory_Load(byte patchnum){
	byte i, v, lsb, msb, in;
	int addr, anal;
	boolean b;
	
	cli();                                      //halt interrupts (i.e. audio engine).  don't want it interferring with eeprom reading
	
	addr = patchnum * MEM_PATCHSIZE;            //calculate start address.  default patch size is 19 bytes
	v = 0;
	
	for (i=0;i<8;i++){                          //set function values
		in = EEPROM.read(addr + i);               //read byte from memory
		lsb = in & B00001111;                     //2 function values in each byte.  first one can be obtained by blanking top 4 bits
		Hardware_Let_Value(v, lsb);
		msb = in >> 4;                            //second one can be obtained by bit-shifting 4 bits to the right
		Hardware_Let_Value(v+1, msb);
		v += 2;
	}
	//set ctrls
	anal = EEPROM.read(addr+8);                //first the 6 red banks ctrls
	Hardware_Let_Ctrl(0, CTRL_FILT, anal);
	anal = EEPROM.read(addr+9);
	Hardware_Let_Ctrl(0, CTRL_Q, anal);
	anal = EEPROM.read(addr+10);
	Hardware_Let_Ctrl(0, CTRL_ENV, anal);
	anal = EEPROM.read(addr+11);
	Hardware_Let_Ctrl(0, CTRL_LFO, anal);
	anal = EEPROM.read(addr+12);
	Hardware_Let_Ctrl(0, CTRL_AMP, anal);
	anal = EEPROM.read(addr+13);
	Hardware_Let_Ctrl(0, CTRL_FX, anal);
	anal = EEPROM.read(addr+14);                //first the 6 red banks ctrls
	Hardware_Let_Ctrl(1, CTRL_FILT, anal);
	anal = EEPROM.read(addr+15);
	Hardware_Let_Ctrl(1, CTRL_Q, anal);
	anal = EEPROM.read(addr+16);
	Hardware_Let_Ctrl(1, CTRL_ENV, anal);      //the the 4 green bank ctrls
	anal = EEPROM.read(addr+17);
	Hardware_Let_Ctrl(1, CTRL_LFO, anal);
	anal = EEPROM.read(addr+18);
	Hardware_Let_Ctrl(1, CTRL_AMP, anal);
	anal = EEPROM.read(addr+19);
	Hardware_Let_Ctrl(1, CTRL_FX, anal);
	
	//set function shift modes.  these are all stored in 1 byte, as each boolean value is only 1 bit.
	in = EEPROM.read(addr+20);                 //read entire byte
	if (bitRead(in, 0)==1) {                   //then just look at relevant bit
		Hard_Let_Shift(FUNC_WAVEA,true);          //and set wavetable bank accordingly
	}
	else{
		Hard_Let_Shift(FUNC_WAVEA,false);
	}
	if (bitRead(in, 1)==1) {                   //filter normalise mode
		Hard_Let_Shift(FUNC_WAVEB,true);
	}
	else{
		Hard_Let_Shift(FUNC_WAVEB,false);
	}
	if (bitRead(in, 2)==1) {
		Hard_Let_Shift(FUNC_FENVA,true);         //filter/pitch env invert mode
	}
	else{
		Hard_Let_Shift(FUNC_FENVA,false);
	}
	if (bitRead(in, 3)==1) {
		Hard_Let_Shift(FUNC_LFOTYPE,true);       //lfo invert mode
	}
	else{
		Hard_Let_Shift(FUNC_LFOTYPE,false);
	}
	if (bitRead(in, 4)==1) {
		Hard_Let_Shift(FUNC_ARPTYPE,true);        //arpeggiator pingpong
	}
	else{
		Hard_Let_Shift(FUNC_ARPTYPE,false);
	}
	if (bitRead(in, 5)==1) {
		Hard_Let_Shift(FUNC_PORTA,true);          //proportional portamento
	}
	else{
		Hard_Let_Shift(FUNC_PORTA,false);
	}
	if (bitRead(in, 6)==1) {
		Hard_Let_Shift(FUNC_BITCRUSH,true);       //filter before bit crusher
	}
	else{
		Hard_Let_Shift(FUNC_BITCRUSH,false);
	}
	
	sei();

}

//***** User wave ******
//save user wave
void Memory_UserWave_Write(byte patchnum){
	int i, addr;
	cli();                                                   //stop interrupts (audio engine)
	addr = ((int)patchnum * WAVE_LEN) + (MEM_PATCHSIZE*16);  //calc start address. 288 bytes is the offset from 16 mem patches
	for (i=0;i<WAVE_LEN;i++){
		EEPROM.write(addr + i, Wave_Get_UserWave(i));          //write each sample to memory
	}
	sei();                                                   //restart interrupts (audio engine)
}

//load user wave
void Memory_UserWave_Read(byte patchnum){
	int i, addr;
	cli();                                                   //stop interrupts (audio engine)
	addr = ((int)patchnum * WAVE_LEN) + (MEM_PATCHSIZE*16);  //calc start address. 288 bytes is the offset from 16 mem patches
	for (i=0;i<WAVE_LEN;i++){
		Wave_Let_UserWave(i,EEPROM.read(addr + i));            //read each sample value from memory
	}
	sei();                                                   //restart interrupts (audio engine)
}

//initialise atmegatron with vanilla patch
void Memory_Vanilla(){
	Hardware_Let_Value(FUNC_WAVEA, 0);        //square wave
	Hardware_Let_Value(FUNC_WAVEB, 0);        //square
	Hardware_Let_Value(FUNC_FENVA, 0);       //'donk' shape filter env
	Hardware_Let_Value(FUNC_FENVDR, 6);
	Hardware_Let_Value(FUNC_FENVS, 0);
	Hardware_Let_Value(FUNC_AENVA, 0);       //full sustain amp env with short decay if sustain lowered
	Hardware_Let_Value(FUNC_AENVD, 6);
	Hardware_Let_Value(FUNC_AENVS, 15);
	Hardware_Let_Value(FUNC_AENVR, 0);
	Hardware_Let_Value(FUNC_LFOTYPE, 0);     //sine LFO
	Hardware_Let_Value(FUNC_LFOSPEED, 8);    //semi-quavers
	Hardware_Let_Value(FUNC_ARPTYPE, 0);     //arp off
	Hardware_Let_Value(FUNC_ARPSPEED, 5);    //quavers
	Hardware_Let_Value(FUNC_PORTA, 0);       //porta off
	Hardware_Let_Value(FUNC_BITCRUSH, 0);    //bitcrush off
	Hardware_Let_Ctrl(0,CTRL_FILT,255);      //filter full open
	Hardware_Let_Ctrl(1,CTRL_FILT,0);    
	Hardware_Let_Ctrl(0,CTRL_Q,0);           //resonance at min (0.5)
	Hardware_Let_Ctrl(1,CTRL_Q,0);        
	Hardware_Let_Ctrl(0,CTRL_ENV,0);         //All Envs and LFO amts = 0
	Hardware_Let_Ctrl(1,CTRL_ENV,0);
	Hardware_Let_Ctrl(0,CTRL_LFO,0);
	Hardware_Let_Ctrl(1,CTRL_LFO,0);
	Hardware_Let_Ctrl(0,CTRL_AMP,0);
	Hardware_Let_Ctrl(1,CTRL_AMP,0);
	Hardware_Let_Ctrl(0,CTRL_FX,0);          //no distortion
	Hardware_Let_Ctrl(1,CTRL_FX,0);          //no phaser
	Hard_Let_Shift(FUNC_WAVEA, false);        //all shift modes off
	Hard_Let_Shift(FUNC_WAVEB, false);
	Hard_Let_Shift(FUNC_FENVA, false);
	Hard_Let_Shift(FUNC_LFOTYPE, false);
	Hard_Let_Shift(FUNC_ARPTYPE, false);
	Hard_Let_Shift(FUNC_PORTA, false);
	Hard_Let_Shift(FUNC_BITCRUSH, false);
}


void Memory_Channel_Write(byte channel){
	EEPROM.write(511, channel);
}
byte Memory_Channel_Read(){
	byte ch;
	ch = EEPROM.read(511);
	if (ch > 15){
		return 0;
	}
	else{
		return ch;
	}
}
