/*
Copyright 2013 Paul Soulsby www.soulsbysynths.com
    This file is part of Atmegatron.

    Atmegatron is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Atmegatron is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Atmegatron.  If not, see <http://www.gnu.org/licenses/>.
*/

// defines for setting and clearing register bits
#ifndef cbi
#define cbi(sfr, bit) (_SFR_BYTE(sfr) &= ~_BV(bit))
#endif
#ifndef sbi
#define sbi(sfr, bit) (_SFR_BYTE(sfr) |= _BV(bit))
#endif
//

void Init_Register() {

  // Set up Timer 2 to do pulse width modulation on the speaker
  // pin.
  pinMode(3, OUTPUT);

   cli();      //pause interrupts
   //use internal clock
   cbi(ASSR,EXCLK);    //disable external clock
   cbi(ASSR,AS2);      //clock timers from internal clock

   //Fast PWM mode
   sbi(TCCR2A,WGM20);  //TOV flag set on max
   cbi(TCCR2A,WGM21);  //Update OCRx at bottom
   cbi(TCCR2B,WGM22);  //Top = 0xFF

   //Set non-inverting PWM on OC2B pin
   cbi(TCCR2A,COM2B0);
   sbi(TCCR2A,COM2B1);
   //set normal operation on OC2A pin
   cbi(TCCR2A,COM2A0);
   cbi(TCCR2A,COM2A1);
   
   //no prescaler
   sbi(TCCR2B,CS20);
   cbi(TCCR2B,CS21);
   cbi(TCCR2B,CS22);
   
   sbi(TIMSK2,TOIE2);
 
 
  // Set Timer 1 to update sample on interrupt


  // Set CTC mode (Clear timer on compare match)
  cbi(TCCR1A,WGM10);
  cbi(TCCR1A,WGM11);
  sbi(TCCR1B,WGM12);
  cbi(TCCR1B,WGM13);

  //No prescaler
  sbi(TCCR1B,CS10);
  cbi(TCCR1B,CS11);
  cbi(TCCR1B,CS12);  

  // Set the compare register (OCR1A) initial value
  OCR1A = F_CPU/(440*WAVE_LEN);       //initialise with A=440hz.

  // Timer 1, output compare A match interrupt enable
  sbi(TIMSK1,OCIE1A);

  // Set Timer 1 to update sample on interrupt


  // Set CTC mode (Clear timer on compare match)
  cbi(TCCR0A,WGM00);
  sbi(TCCR0A,WGM01);
  cbi(TCCR0B,WGM02);


  //prescaler
  cbi(TCCR0B,CS00);
  sbi(TCCR0B,CS01);
  cbi(TCCR0B,CS02);

  // Set the compare register (OCR1A) initial value
  OCR0A = (F_CPU/(440*WAVE_LEN)) >> 2;       //initialise with A=440hz.

  // Timer 1, output compare A match interrupt enable
  sbi(TIMSK0,OCIE0A);


  sei();    //start interrupts
  
  // faster ADC
  // uses div factor of 8.  standard arduino read is 128.
  sbi(ADCSRA,ADPS2) ;
  cbi(ADCSRA,ADPS1) ;
  sbi(ADCSRA,ADPS0) ;
  
}
